/*
 *                            COPYRIGHT
 *
 *  cschem - modular/flexible schematics editor - lib: compile helpers
 *  Copyright (C) 2022 Tibor 'Igor2' Palinkas
 *
 *  (Supported by NLnet NGI0 PET Fund in 2022)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.*
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/cschem
 *    lead developer: email to cschem (at) igor2.repo.hu
 *    mailing list: cschem (at) list.repo.hu (send "subscribe")
 */

#include "config.h"

#include <librnd/core/error.h>

#include "abstract.h"
#include "compile.h"

#include "util_compile.h"

csch_anet_t *csch_cmp_merge_nets(csch_anet_t *n1, csch_anet_t *n2)
{
	csch_anet_t *src, *dst;
	long n;

	/* dst should be the one without user chosen name */
	if (!n1->no_uname && !n2->no_uname) {
		rnd_message(RND_MSG_ERROR, "csch_cmp_merge_nets(): different networks '%s' and '%s' can not be connected\n", n1->name, n2->name);
		return NULL;
	}
	if (!n1->no_uname) {
		dst = n1;
		src = n2;
	}
	else {
		dst = n2;
		src = n1;
	}

	/* switch over the concrete->abstract pointers on src net so src concrete
	   objects know they belong to dst net now */
	for(n = 0; n < src->hdr.srcs.used; n++) {
		csch_cgrp_t *cgrp = src->hdr.srcs.array[n];
		long i;

		if (!csch_obj_is_grp(&cgrp->hdr))
			continue;

		for(i = 0; i < cgrp->aid.used; i++)
			if (cgrp->aid.array[i] == src->hdr.aid)
				cgrp->aid.array[i] = dst->hdr.aid;
	}

	for(n = 0; n < src->conns.used; n++) {
		csch_aport_t *ap = (csch_aport_t *)src->conns.array[n];
		if (ap->hdr.type == CSCH_ATYPE_PORT)
			ap->conn.net = NULL; /* disconnect from src */
		csch_compile_connect_net_to(&dst, &ap->hdr, 1);
	}
	src->conns.used = 0;

	return dst;
}

int csch_cmp_nameconn_port_net(csch_acomp_t *comp, const char *portname, const char *netname)
{
	csch_aport_t *port = csch_aport_get(comp->hdr.abst, comp, portname, 1);
	csch_anet_t *net; /* net from the argument */

	if (port == NULL) {
		rnd_message(RND_MSG_ERROR, "Component '%s' doesn't contain port named '%s'\n", comp->name, portname);
		return -1;
	}

	net = csch_anet_get(comp->hdr.abst, netname, 1, 0);

	if (net == NULL) {
		rnd_message(RND_MSG_ERROR, "Netname '%s' does not exist\n", netname);
		return -1;
	}
	if (port->conn.net == NULL) {
		rnd_message(RND_MSG_ERROR, "Component '%s' port '%s' has empty conn list\n", comp->name, portname);
		return -1;
	}

	return (csch_cmp_merge_nets(net, port->conn.net) != NULL) ? 0 : -1;
}

csch_inline int cmp_attr_derive_scalar(csch_ahdr_t *aobj, const char *dkey, const char *skey, int prio, csch_source_arg_t src, csch_attr_derive_t how)
{
	vtp0_t *srcs = &aobj->srcs;
	int op = how & CSCH_CMPDRV_opmask;
	long n;
	gds_t tmp = {0};

	for(n = 0; n < srcs->used; n++) {
		csch_cgrp_t *cgrp = srcs->array[n];
		const char *s = csch_attrib_get_str(&cgrp->attr, skey);

		if (s == NULL) continue;

		if (tmp.used != 0) {
			switch(op) {
				case CSCH_CMPDRV_UNIQ:
					if (strcmp(tmp.array, s) != 0) {
						rnd_message(RND_MSG_ERROR, "Multiple incompatible values for key '%s'\n", skey);
						goto error;
					}
					break;
				case CSCH_CMPDRV_CONCAT:
					gds_append(&tmp, ',');
					gds_append_str(&tmp, s);
					break;
			}
		}
		else
			gds_append_str(&tmp, s);
	}

	csch_attrib_set(&aobj->attr, prio, dkey, tmp.array, src, NULL);

	gds_uninit(&tmp);
	return 0;

	error:;
	gds_uninit(&tmp);
	return -1;
}

csch_inline int cmp_attr_derive_arr(csch_ahdr_t *aobj, const char *dkey, const char *skey, int prio, csch_source_arg_t src, csch_attr_derive_t how)
{
	TODO("Implement this when needed");
	return -1;
}

int csch_cmp_attr_derive(csch_ahdr_t *aobj, const char *dkey, const char *skey, int prio, csch_source_arg_t src, csch_attr_derive_t how)
{
	vtp0_t *srcs = &aobj->srcs;
	long n;
	int is_str = 0, is_arr = 0;

	/* map attr types */
	for(n = 0; n < srcs->used; n++) {
		csch_cgrp_t *cgrp = srcs->array[n];
		csch_attrib_t *a = csch_attrib_get(&cgrp->attr, skey);

		if (a == NULL) continue;

		if (a->val != NULL) is_str++;
		if (a->arr.used != 0) is_arr++;
	}

	/* tail-recurse to per type implementation */
	if (is_str) return cmp_attr_derive_scalar(aobj, dkey, skey, prio, src, how);
	if (is_arr) return cmp_attr_derive_arr(aobj, dkey, skey, prio, src, how);

	/* empty result */
	if (how & CSCH_CMPDRV_EMPTY_SKIP)
		return 0;
	if (how & CSCH_CMPDRV_EMPTY_DEL)
		return csch_attrib_del(&aobj->attr, prio, dkey, src);

	return 0; /* default is skip */
}

